"""Data Modules."""

from .generic_classification_module import (  # noqa: F401
    OlindaCDataModule,
    OlindaCDataset,
)
from .generic_regression_module import OlindaRDataModule, OlindaRDataset  # noqa: F401
